#!/usr/bin/env python3

'''Convert tagAlign files for further processing.'''

import os
import argparse
import shutil
import subprocess
import shlex
import logging
from multiprocessing import cpu_count
import utils

EPILOG = '''
For more details:
        %(prog)s --help
'''

# SETTINGS

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
logger.propagate = False
logger.setLevel(logging.INFO)


def get_args():
    '''Define arguments.'''

    parser = argparse.ArgumentParser(
        description=__doc__, epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-b', '--bam',
                        help="The bam file to convert.",
                        required=True)

    parser.add_argument('-p', '--paired',
                        help="True/False if paired-end or single end.",
                        default=False,
                        action='store_true')

    parser.add_argument('-m', '--chrm',
                        help="True/False if to remove ChrM.",
                        default=False,
                        action='store_true')

    parser.add_argument('-t', '--tn5',
                    help='Enable TN5 shifting for ATAC-Seq.',
                    default=False,
                    action='store_true')

    args = parser.parse_args()
    return args

# Functions


def check_tools():
    '''Checks for required componenets on user system'''

    logger.info('Checking for required libraries and components on this system')

    bedtools_path = shutil.which("bedtools")
    if bedtools_path:
        logger.info('Found bedtools: %s', bedtools_path)
    else:
        logger.error('Missing bedtools')
        raise Exception('Missing bedtools')

    samtools_path = shutil.which("samtools")
    if samtools_path:
        logger.info('Found samtools: %s', samtools_path)
    else:
        logger.error('Missing samtools')
        raise Exception('Missing samtools')


def convert_mapped(bam, tag_filename, chrm):
    '''Use bedtools to convert to tagAlign.'''

    steps = [
        "bamToBed -i %s" % (bam),
        r"""awk 'BEGIN{OFS="\t"}{$4="N";$5="1000";print $0}'"""]

    if chrm:
        steps.extend(["grep -v 'chrM'"])

    steps.extend(["gzip -nc"])

    out, err = utils.run_pipe(steps, outfile=tag_filename)


def convert_mapped_pe(bam, bam_basename, tag_filename, chrm):
    '''Use bedtools to convert to tagAlign PE data.'''

    bedpe_filename = bam_basename + ".bedpe.gz"

    # Name sort bam to make BEDPE
    nmsrt_bam_filename = bam_basename + ".nmsrt.bam"
    samtools_sort_command = \
        "samtools sort -n -@%d -o %s %s" \
        % (cpu_count(), nmsrt_bam_filename, bam)

    logger.info(samtools_sort_command)
    subprocess.check_output(shlex.split(samtools_sort_command))

    steps = ["bamToBed -bedpe -mate1 -i %s" % (nmsrt_bam_filename)]

    steps.extend(["gzip -nc"])

    out, err = utils.run_pipe(steps, outfile=bedpe_filename)
    os.remove(nmsrt_bam_filename)

    if chrm:
        steps.extend(["grep -v 'chrM'"])

    # Convert read pairs to reads into standard tagAlign file
    tag_steps = ["zcat -f %s" % (bedpe_filename)]
    tag_steps.extend([r"""awk 'BEGIN{OFS="\t"}{printf "%s\t%s\t%s\tN\t1000\t%s\n%s\t%s\t%s\tN\t1000\t%s\n",$1,$2,$3,$9,$4,$5,$6,$10}'"""])
    if chrm:
        tag_steps.extend(["grep -v 'chrM'"])
    tag_steps.extend(['gzip -cn'])
    out, err = utils.run_pipe(tag_steps, outfile=tag_filename)
    os.remove(bedpe_filename)


def tn5_shift_tagalign(tag_filename, shitfted_tag_filename):
    '''Shift tagalign file for Tn5.'''

    steps = [
        "zcat -f %s" % (tag_filename),
        r"""awk 'BEGIN {OFS = "\t"}{ if ($6 == "+") {$2 = $2 + 4} else if ($6 == "-") {$3 = $3 - 5} print $0}'""",
        "gzip -nc"]
    out, err = utils.run_pipe(steps, outfile=shitfted_tag_filename)


def main():
    args = get_args()
    paired = args.paired
    bam = args.bam
    chrm = args.chrm
    tn5 = args.tn5

    # Create a file handler
    handler = logging.FileHandler('convert_reads.log')
    logger.addHandler(handler)

    # Check if tools are present
    check_tools()

    # Convert PE or SE to tagAlign
    bam_basename = os.path.basename(
        utils.strip_extensions(bam, ['.bam']))

    tag_filename = bam_basename + '.tagAlign.gz'

    shitfted_tag_filename = bam_basename + ".tn5.tagAlign.gz"


    if paired:  # paired-end data
        convert_mapped_pe(bam, bam_basename, tag_filename, chrm)
    else:
        convert_mapped(bam, tag_filename, chrm)

    if tn5:
        logger.info("TN5-shifting TAGALIGN...")
        tn5_shift_tagalign(tag_filename, shitfted_tag_filename)
    else:
        tn5_shift = ["cp %s %s" % (tag_filename, shitfted_tag_filename)]
        out, err = utils.run_pipe(tn5_shift)

    os.remove(tag_filename)

if __name__ == '__main__':
    main()
